/*jslint indent:4,node:true,nomen:true*/
/*global describe,before,it,after*/
(function () {

    'use strict';

    var _ = require('underscore'),
        sinon = require('sinon'),
        expect = require('expect.js'),
        utils = require('./utils'),
        perl = require('../lib/perl'),
        path = require('path'),
        factory = require('../lib/argparse-factory'),
        stub;

    describe('Checking argparser-factory.js calling', function () {
        before(function() {
            stub = sinon.stub(perl, 'get');

            stub.withArgs('env', ['webroot', 'log', 'version']).yields(null, {
                webroot: path.join(__dirname, 'fs/webroot'),
                version: '6.xx.x'
            });

            stub.withArgs('env', ['webroot', 'log']).yields(null, {
                webroot: path.join(__dirname, 'fs/webroot')
            });

            stub.withArgs('cartridges').yields(null, [
                'PARENT/JQuery'
            ]);

            stub.withArgs('uid').yields(42);
            stub.withArgs('gid').yields(1337);

            stub.withArgs('env', ['cartridges']).yields(null, {cartridges: path.join(__dirname, 'fs/cartridges')});
        });

        after(function () {
            perl.get.restore();
        });

        // Examplary test for checking error handling (1) and module call (2)
        var moduleToCheck = {
            'pack-lib': {
                mocked: {
                    argsToFail: ['', 'epnode6-test', 'pack-lib', '-s', 'This_does_not_exist'],
                    argsToPass: ['', 'epnode6-test', 'pack-lib', '-s', 'MyStore']
                },
                expectedCallArgs: {
                    'subcommand': 'pack-lib',
                    'store_type': 'MyStore',
                }
            },
            'minify-lib': {
                mocked: {
                    argsToFail: ['', 'epnode6-test', 'minify-lib', '-s', 'This_does_not_exist'],
                    argsToPass: ['', 'epnode6-test', '-l', '0', 'minify-lib', '-s', 'MyStore', '-f', '"**/*.ca.json"']
                },
                expectedCallArgs: {
                    'log_level': '0',
                    'subcommand': 'minify-lib',
                    'store_type': 'MyStore',
                    'filter': '"**/*.ca.json"'
                }
            },
            'minify': {
                mocked: {
                    argsToFail: ['', 'epnode6-test', 'minify', '-s', 'This_does_not_exist', '-t', 'This_does_not_exist_either'],
                    argsToPass: ['', 'epnode6-test', 'minify', '-s', 'The_Source', '-t', 'The_Target']
                },
                expectedCallArgs: {
                    'subcommand': 'minify',
                    'source': 'The_Source',
                    'target': 'The_Target'
                }
            },
            'build-ui': {
                mocked: {
                    argsToFail: ['', 'epnode6-test', 'build-ui', '-s', 'This_does_not_exist'],
                    argsToPass: ['', 'epnode6-test', 'build-ui', '-s', 'MyStore', '-f', '"**/*.ca.json"']
                },
                expectedCallArgs: {
                    'subcommand': 'build-ui',
                    'store_type': 'MyStore',
                    'filter': '"**/*.ca.json"'
                }
            },
            'testcases': {
                mocked: {
                    argsToFail: ['', 'epnode6-test', 'testcases', '-c', 'PARENT/This_cartridge_does_not_exist'],
                    argsToPass: ['', 'epnode6-test', 'testcases', '-c', 'PARENT/JQuery']
                },
                expectedCallArgs: {
                    'subcommand': 'testcases',
                    'cartridge': 'PARENT/JQuery'
                }
            }
        };
        _.each(moduleToCheck, function (module, moduleName) {
            describe('module ' + moduleName, function () {
                // (1) Check error handling with passing wrong arguments to the module
                describe('with wrong arguments', function () {
                    before(function () {
                        process.argv = module.mocked.argsToFail;
                    });

                    it('should return an error (non-existent path)', function (callback) {
                        factory.modules[moduleName](utils.joinArgs(factory.parser.parseArgs()), function (error) {
                            // We do not determine if we're thrown the correct error since the path on win32 differs from unix.
                            expect(error).not.to.be.empty();
                            callback();
                        });
                    });
                });

                // (2) We are not interested in what the function does, but how it is called. Therefore our stub immediately invokes the callback (callsArgs1).
                describe('exec', function () {
                    var stub;

                    before(function () {
                        process.argv = module.mocked.argsToPass;
                        stub = sinon.stub(factory.modules, moduleName).callsArg(1);
                    });

                    it('should call actual module with correct parameters', function (callback) {
                        factory.modules[moduleName](utils.joinArgs(factory.parser.parseArgs()), function (error) {
                            expect(error).to.be(undefined);
                            expect(stub.getCall(0).args[0]).to.eql(module.expectedCallArgs);

                            callback();
                        });
                    });

                    after(function () {
                        stub.restore();
                    });
                });
            });
        });
    });
}());