/*jslint node:true, nomen:true*/
/*global describe,before,it,after*/

(function () {

    'use strict';

    var _ = require('underscore'),
        async = require('async'),
        expect = require('expect.js'),
        fs = require('fs'),
        grunt = require('grunt'),
        sinon = require('sinon'),
        minify = require('../lib/minify-batch'),
        path = require('path'),
        os = require('os'),
        perl = require('../lib/perl'),

        is_unixWithRoot = (os.platform() !== 'win32' && process.getuid() === 0) ? 'Running on Unix with root, running minify test with checks for uid and gid' : undefined;

    describe('Running minify-batch.js', function () {
        var webrootLib = path.join(__dirname, 'fs/webroot/StoreTypes/6.xx.x/MyStore/lib'),
            webrootLibMin = path.join(__dirname, 'fs/webroot/StoreTypes/6.xx.x/MyStore/lib-min'),
            expectedPath = path.join(__dirname, 'fs/expected'),
            pageCache = path.join(__dirname, 'fs/pageCache'),
            pageCacheMin = path.join(__dirname, 'fs/pageCache-min'),
            options = {};

        // invalidate require cache because we overwrote some arguments in the test for the argparser factory
        require.cache[path.normalize(__dirname + '/../lib/minify-batch.js')] = undefined;

        before(function () {
            if (is_unixWithRoot) {
                options.uid = 1337;
                options.gid = 42;
            }

            // mock the perl module
            sinon.stub(perl, 'get').withArgs('cartridges', 'Site').yields(null, [
                'PARENT/JQuery',
                'PARENT/CartridgeD',
                'PARENT/CartridgeE',
                'PARENT/CartridgeB',
                'Another_Parent/CartridgeC',
                'PARENT/CartridgeA'
            ]);
        });

        after(function () {
            perl.get.restore();
        });

        describe('with filter option', function () {
            before(function (callback) {
                var optionsWithFilter = _.extend({}, options, { filter: '**/*.klingon.json' });

                minify(webrootLib, webrootLibMin, optionsWithFilter, callback);
            });

            after(function () {
                if (grunt.file.isDir(webrootLibMin)) {
                    grunt.file['delete'](webrootLibMin);
                }
            });

            it('should copy the klingon dictionary but omit the zulu one according to the provided filter', function () {
                expect(grunt.file.isFile(path.join(webrootLibMin, 'ep/dict.klingon.json'))).to.be.ok();
                expect(grunt.file.isFile(path.join(webrootLibMin, 'ep/dict.zulu.json'))).not.to.be.ok();
            });
        });

        describe('to process the page cache', function () {
            before(function (callback) {
                minify(pageCache, pageCacheMin, options, callback);
            });

            after(function () {
                if (grunt.file.isDir(pageCacheMin)) {
                    grunt.file['delete'](pageCacheMin);
                }
            });

            it('should minify *.content files correctly', function (callback) {
                async.series({
                    expectedStartpage: function (callback) {
                        fs.readFile(path.join(expectedPath, 'startpage.compressed.content'), callback);
                    },

                    expectedProductpage: function (callback) {
                        fs.readFile(path.join(expectedPath, 'productpage.compressed.content'), callback);
                    },

                    actualStartpage: function (callback) {
                        fs.readFile(path.join(pageCacheMin, 'startpage.content'), callback);
                    },

                    actualProductpage: function (callback) {
                        fs.readFile(path.join(pageCacheMin, 'productpage.content'), callback);
                    }
                }, function (error, result) {
                    expect(error).not.to.be.ok();
                    expect(result.actualStartpage.toString()).to.be(result.expectedStartpage.toString());
                    expect(result.actualProductpage.toString()).to.be(result.expectedProductpage.toString());

                    callback();
                });
            });
        });

        describe('recursively for all files', function () {
            before(function (callback) {
                minify(webrootLib, webrootLibMin, options, callback);
            });

            after(function () {
                if (grunt.file.isDir(webrootLibMin)) {
                    grunt.file['delete'](webrootLibMin);
                }
            });

            it('should minify a2.js.map correctly', function (callback) {
                async.parallel({
                    expected: function (callback) {
                        fs.readFile(path.join(expectedPath, 'parent-a2.min.js.map'), callback);
                    },

                    actual: function (callback) {
                        fs.readFile(path.join(webrootLibMin, 'parent', 'a2.js.map'), callback);
                    }
                }, function (error, result) {
                    expect(error).to.not.be.ok();
                    expect(result.expected.toString()).to.be(result.actual.toString());

                    // if we're root, we check if file ownership is set correctly
                    if (is_unixWithRoot) {
                        fs.stat(path.join(webrootLibMin, 'parent', 'a2.js.map'), function (err, stats) {
                            expect(err).to.not.be.ok();

                            expect(stats.uid).to.be.a('number');
                            expect(stats.uid).to.be(1337);
                            expect(stats.gid).to.be.a('number');
                            expect(stats.gid).to.be(42);

                            expect(parseInt(stats.mode.toString(8), 10)).to.be(100640);

                            callback();
                        });
                    } else {
                        callback();
                    }
                });
            });

            it('should minify package-bo.less and place package-bo.css into lib-min', function (callback) {
                async.parallel({
                    expected: function (callback) {
                        fs.readFile(path.join(expectedPath, 'package-bo.min.css'), callback);
                    },

                    actual: function (callback) {
                        fs.readFile(path.join(webrootLibMin, 'package-tominify.css'), callback);
                    }
                }, function (error, result) {
                    expect(error).to.not.be.ok();
                    // replace CR (\r) and LF (\n) by LF (\n) due to different new line handling in different operating systems
                    expect(result.expected.toString().replace(/\r\n/g, "\n")).to.be(result.actual.toString().replace(/\r\n/g, "\n"));

                    // if we're root, we check if file ownership is set correctly
                    if (is_unixWithRoot) {
                        fs.stat(path.join(webrootLibMin, 'package-tominify.css'), function (err, stats) {
                            expect(err).to.not.be.ok();

                            expect(stats.uid).to.be.a('number');
                            expect(stats.uid).to.be(1337);
                            expect(stats.gid).to.be.a('number');
                            expect(stats.gid).to.be(42);

                            expect(parseInt(stats.mode.toString(8), 10)).to.be(100640);

                            callback();
                        });
                    } else {
                        callback();
                    }
                });
            });

            it('should not place template files within the minified directory', function () {
                expect(grunt.file.isFile(path.join(webrootLibMin, 'backbone-someview.tmpl.html'))).not.to.be.ok();
            });

            it('should not place compiled templates within the minified directory', function () {
                expect(grunt.file.isFile(path.join(webrootLib, 'backbone-someview.tmpl.js'))).not.to.be.ok();
            });
        });
    });
}());