/*jslint node:true,nomen:true*/
/*global describe,before,it,after*/
(function () {

    'use strict';

    var expect = require('expect.js'),
        sinon  = require('sinon'),
        path = require('path'),
        child_process = require('child_process'),
        commons = require('../lib/commons');

    describe('Running perl.js', function () {
        var epagesPerlOriginal,
            epagesVersionOriginal,
            epagesWebrootOriginal,
            perl,
            perlLocation = path.normalize(path.join(__dirname, '../lib/', 'perl')),
            lastPerlCmd;

        before(function () {
            // define envs
            epagesPerlOriginal = process.env.EPAGES_PERL;
            process.env.EPAGES_PERL = '__PERL__';

            // remove perl and child_process module from cache
            require.cache[perlLocation + '.js'] = undefined;
            require.cache.child_process = undefined;

            // mock child_process.exec
            sinon.stub(child_process, 'exec', function (cmd, callback) {
                var normalizedCmd = [],
                    params,
                    interpreterAndPath;

                // split command into interpreter + script and parameters
                cmd = cmd.split(/\s(-.*)/);
                interpreterAndPath = cmd[0].split(' ');
                params = cmd[1].trim();

                // interpreter
                normalizedCmd[0] = interpreterAndPath[0];
                // actual script path
                // `__dirname` might have spaces too so we stich the path back together.
                normalizedCmd[1] = interpreterAndPath.length > 2 ? interpreterAndPath.splice(1, interpreterAndPath.length - 1).join(' ') : interpreterAndPath[1];
                // parameters
                normalizedCmd[2] = params;

                lastPerlCmd = normalizedCmd;

                switch (normalizedCmd[2]) {
                case '-sub "GetStoreTypeCartridges" -param "Store"':
                    callback(null, '["DE_EPAGES::One", "DE_EPAGES::Two"]');
                    break;
                case '-sub "GetStoreTypeCartridges" -param "Site"':
                    callback(null, '["DE_EPAGES::Eleven", "DE_EPAGES::Twelve"]');
                    break;
                case '-sub "GetVersion"':
                    callback(null, '"17.2.6"');
                    break;
                default:
                    callback('This behaviour is not mocked.');
                    break;
                }
            });

            perl = require(perlLocation);

        });

        after(function (callback) {
            // restore envs
            process.env.EPAGES_PERL = epagesPerlOriginal;
            process.env.EPAGES_VERSION = epagesVersionOriginal;
            process.env.EPAGES_WEBROOT = epagesWebrootOriginal;

            // restore child process module
            child_process.exec.restore();

            callback();
        });

        it('should return an error if an unknown getter is requested', function (callback) {
            perl.get('iGuessThisGetterDoesntExist', function (error) {
                expect(error).to.be('Unknown getter iGuessThisGetterDoesntExist');
                callback();
            });
        });

        describe('get cartridges', function () {
            it('should run the correct external command ([Interpreter] [ScriptPath}] [Params]', function (callback) {
                perl.get('cartridges', function () {
                    expect(lastPerlCmd).to.be.an('array');

                    expect(lastPerlCmd[0]).to.be('__PERL__/bin/perl');
                    // join something like: someFolder/lib/../perlLocation to someFolder/perlLocation
                    expect(commons.fixPath(path.join(lastPerlCmd[1]))).to.be(
                        commons.fixPath(path.normalize(path.join(__dirname, '..', 'child_process', 'jsapi.pl')))
                    );
                    expect(lastPerlCmd[2]).to.be('-sub "GetStoreTypeCartridges" -param "Store"');

                    callback();
                });
            });

            it('should return the cartridges for Store if no store type is given', function (callback) {
                perl.get('cartridges', function (error, result) {
                    expect(error).to.not.be.ok();

                    expect(result).to.be.an('array');

                    expect(result[0]).to.be('DE_EPAGES/One');
                    expect(result[1]).to.be('DE_EPAGES/Two');

                    callback();
                });
            });

            it('should return the cartridges for Site if the store type is set to Site', function (callback) {
                perl.get('cartridges', 'Site', function (error, result) {
                    expect(error).to.not.be.ok();

                    expect(result).to.be.an('array');

                    expect(result[0]).to.be('DE_EPAGES/Eleven');
                    expect(result[1]).to.be('DE_EPAGES/Twelve');

                    callback();
                });
            });
        });

        describe('get environment variables', function () {
            before(function (callback) {
                epagesWebrootOriginal = process.env.EPAGES_WEBROOT;
                process.env.EPAGES_WEBROOT = '/this/is/our/web root';

                epagesVersionOriginal = process.env.EPAGES_VERSION;
                process.env.EPAGES_VERSION = '17.2.6';

                epagesVersionOriginal = process.env.EPAGES_WEBGROUP;
                process.env.EPAGES_WEBGROUP = 'somegroup:815';

                callback();
            });

            it('should call the external command for the version', function (callback) {
                perl.get('env', function () {

                    expect(lastPerlCmd).to.be.an('object');

                    expect(lastPerlCmd[0]).to.be('__PERL__/bin/perl');
                    // join something like: someFolder/lib/../perlLocation to someFolder/perlLocation
                    expect(commons.fixPath(path.join(lastPerlCmd[1]))).to.be(
                        commons.fixPath(path.normalize(path.join(__dirname, '../lib/../', 'child_process', 'jsapi.pl')))
                    );
                    expect(lastPerlCmd[2]).to.be('-sub "GetVersion"');

                    callback();
                });
            });

            it('should return the webroot', function (callback) {
                perl.get('env', ['webroot'], function (error, result) {
                    expect(error).to.not.be.ok();

                    expect(lastPerlCmd).to.be.an('object');
                    expect(result.webroot).to.be('/this/is/our/web root');

                    callback();
                });
            });

            it('should return EPAGES_VERSION if no variable is given', function (callback) {
                perl.get('env', function (error, result) {
                    expect(error).to.not.be.ok();

                    expect(result).to.be.an('object');
                    expect(result.version).to.be('17.2.6');

                    callback();
                });
            });

            it('should return correct gid of EPAGES_WEBGROUP', function (callback) {
                perl.get('gid', 'webgroup', function (result) {

                    expect(result).to.be.an('number');
                    expect(result).to.be(815);

                    callback();
                });
            });
        });
    });

}());
